@@ -1,48 +1,178 @@
 // Copyright 2015 Ted Mielczarek. See the COPYRIGHT
 // file at the top-level directory of this distribution.
 
-use chrono::{TimeZone, Utc};
-use failure::Fail;
-
+use std::collections::{BTreeMap, BTreeSet};
 use std::ops::Deref;
 use std::path::Path;
+use std::time::{Duration, SystemTime};
 
 use minidump::{self, *};
 
+use crate::arg_recovery;
 use crate::evil;
 use crate::process_state::{CallStack, CallStackInfo, LinuxStandardBase, ProcessState};
 use crate::stackwalker;
 use crate::symbols::*;
 use crate::system_info::SystemInfo;
 
-/// Various advanced options for the processor.
-#[derive(Default, Debug, Clone)]
+/// Configuration of the processor's exact behaviour.
+///
+/// This can be used to either:
+///
+/// * enable extra features that are disabled by default
+/// * lock in the features you want enabled to minimize future changes
+///
+/// All fields are `pub`, but the type is `non_exhaustive`.
+/// Recommended usage is to call one of the constructors to get a baseline
+/// set of features, and then manually set any values you particularly care about.
+///
+/// If we decide an unstable feature exposed by these flags is a bad idea,
+/// we may remove its functionality and turn it into a noop, but the flag
+/// will remain to avoid breaking code. Similarly, if a feature seems to be
+/// too bloated, its implementation may be hidden behind a cargo feature
+/// flag, producing a similar result if that feature is statically disabled.
+///
+/// In either of these cases, a `warn` diagnostic will be emitted if you
+/// try to use request a feature whose implementation does not exist.
+///
+/// [`process_minidump`][] uses [`ProcessorOptions::stable_basic`][], which
+/// is also exposed as [`Default::default`].
+///
+/// ## Example:
+///
+/// ```
+/// use minidump_processor::ProcessorOptions;
+///
+/// // Happy with the default-enabled features
+/// let mut options = ProcessorOptions::stable_basic();
+/// // But specifically want this cool unstable feature
+/// options.recover_function_args = true;
+/// ```
+///
+#[derive(Debug, Clone)]
 #[non_exhaustive]
 pub struct ProcessorOptions<'a> {
-    /// The evil "raw json" mozilla's legacy infrastructure relies on (to be phased out).
+    /// **\[UNSTABLE\]** The evil "raw json" mozilla's legacy infrastructure relies on.
+    ///
+    /// Please don't use this. If you have to use this, you know who you are.
     pub evil_json: Option<&'a Path>,
+
+    /// **\[UNSTABLE\]** Whether to try to heuristically recover function arguments in backtraces.
+    ///
+    /// Currently this only work for x86, and assumes everything is either cdecl or thiscall
+    /// (inferred from whether the symbol name looks like a static function or a method).
+    pub recover_function_args: bool,
 }
 
-/// An error encountered during minidump processing.
-#[derive(Debug, Fail)]
-pub enum ProcessError {
-    #[fail(display = "Failed to read minidump")]
-    MinidumpReadError(minidump::Error),
-    #[fail(display = "An unknown error occurred")]
-    UnknownError,
-    #[fail(display = "The system information stream was not found")]
-    MissingSystemInfo,
-    #[fail(display = "The thread list stream was not found")]
-    MissingThreadList,
-}
+impl ProcessorOptions<'_> {
+    /// "Do the normal stuff everyone should want"
+    ///
+    /// * `evil_json: None`
+    /// * `recover_function_args: false`
+    ///
+    /// Unlike stable_all, you shouldn't expect this to change its results much.
+    ///
+    /// It will specifically always try to:
+    ///
+    /// * Perform full backtraces and symbolication of every thread.
+    /// * Produce detailed system info (OS, Cpu, Versions...)
+    /// * Produce detailed crash info (Crashing thread, crash address, formatted error...)
+    /// * List loaded and unloaded modules
+    pub fn stable_basic() -> Self {
+        ProcessorOptions {
+            evil_json: None,
+            recover_function_args: false,
+        }
+    }
 
-impl From<minidump::Error> for ProcessError {
-    fn from(err: minidump::Error) -> ProcessError {
-        ProcessError::MinidumpReadError(err)
+    /// "Turn all the stable features on"
+    ///
+    /// * `evil_json: None`
+    /// * `recover_function_args: false`
+    ///
+    /// (At this precise moment this is identical to stable_basic, but may diverge
+    /// as we introduce more features.)
+    ///
+    /// Everything included by stable_basic, but willing to enable more interesting
+    /// features and spend extra time trying to find extra insights. This is the default
+    /// place that unstable features will "graduate" to when they're deemed good enough.
+    pub fn stable_all() -> Self {
+        ProcessorOptions {
+            evil_json: None,
+            recover_function_args: false,
+        }
+    }
+
+    /// "Turn EVERYTHING on, even the experimental stuff!"
+    ///
+    /// * `evil_json: None`
+    /// * `recover_function_args: true`
+    ///
+    /// (evil_json is still "disabled" because you need to give it needs a path.)
+    ///
+    /// Some of this stuff can be really jank, use at your own risk!
+    pub fn unstable_all() -> Self {
+        ProcessorOptions {
+            evil_json: None,
+            recover_function_args: true,
+        }
+    }
+
+    /// Check if any of the enabled features are deprecated or disabled
+    /// and emit warnings if they are.
+    fn check_deprecated_and_disabled(&self) {
+        // Currently nothing is deprecated / disableable, but here's the template.
+
+        /*
+        use log::warn;
+
+        if self.my_bad_feature {
+            warn!("Deprecated ProcessorOption my_bad_feature has been removed and does nothing.")
+        }
+
+        if !cfg!(feature = "my-optional-feature") && self.my_optional_feature {
+            warn!("Disabled ProcessorOption my_optional_feature must be enabled via cargo.")
+        }
+        */
     }
 }
 
-/// Unwind all threads in `dump` and return a `ProcessState`.
+impl Default for ProcessorOptions<'_> {
+    fn default() -> Self {
+        Self::stable_basic()
+    }
+}
+
+/// An error encountered during minidump processing.
+#[derive(Debug, thiserror::Error)]
+pub enum ProcessError {
+    #[error("Failed to read minidump")]
+    MinidumpReadError(#[from] minidump::Error),
+    #[error("An unknown error occurred")]
+    UnknownError,
+    #[error("The system information stream was not found")]
+    MissingSystemInfo,
+    #[error("The thread list stream was not found")]
+    MissingThreadList,
+}
+
+impl ProcessError {
+    /// Returns just the name of the error, as a more human-friendly version of
+    /// an error-code for error logging.
+    pub fn name(&self) -> &'static str {
+        match self {
+            ProcessError::MinidumpReadError(_) => "MinidumpReadError",
+            ProcessError::UnknownError => "UnknownError",
+            ProcessError::MissingSystemInfo => "MissingSystemInfo",
+            ProcessError::MissingThreadList => "MissingThreadList",
+        }
+    }
+}
+
+/// Unwind all threads in `dump` and return a report as a `ProcessState`.
+///
+/// This is equivalent to [`process_minidump_with_options`] with
+/// [`ProcessorOptions::stable_basic`][].
 ///
 /// # Examples
 ///
@@ -50,41 +180,47 @@ impl From<minidump::Error> for ProcessError {
 /// use minidump::Minidump;
 /// use std::path::PathBuf;
 /// use breakpad_symbols::{Symbolizer, SimpleSymbolSupplier};
+/// use minidump_processor::ProcessError;
 ///
-/// # std::env::set_current_dir(env!("CARGO_MANIFEST_DIR"));
-/// # fn foo() -> Result<(), minidump_processor::ProcessError> {
-/// let mut dump = Minidump::read_path("../testdata/test.dmp")?;
-/// let supplier = SimpleSymbolSupplier::new(vec!(PathBuf::from("../testdata/symbols")));
-/// let symbolizer = Symbolizer::new(supplier);
-/// let state = minidump_processor::process_minidump(&mut dump, &symbolizer)?;
-/// assert_eq!(state.threads.len(), 2);
-/// println!("Processed {} threads", state.threads.len());
-/// # Ok(())
-/// # }
-/// # fn main() { foo().unwrap() }
+/// #[tokio::main]
+/// async fn main() -> Result<(), ProcessError> {
+///     # std::env::set_current_dir(env!("CARGO_MANIFEST_DIR"));
+///     let mut dump = Minidump::read_path("../testdata/test.dmp")?;
+///     let supplier = SimpleSymbolSupplier::new(vec!(PathBuf::from("../testdata/symbols")));
+///     let symbolizer = Symbolizer::new(supplier);
+///     let state = minidump_processor::process_minidump(&mut dump, &symbolizer).await?;
+///     assert_eq!(state.threads.len(), 2);
+///     println!("Processed {} threads", state.threads.len());
+///     Ok(())
+/// }
 /// ```
-pub fn process_minidump<'a, T, P>(
+pub async fn process_minidump<'a, T, P>(
     dump: &Minidump<'a, T>,
     symbol_provider: &P,
 ) -> Result<ProcessState, ProcessError>
 where
     T: Deref<Target = [u8]> + 'a,
-    P: SymbolProvider,
+    P: SymbolProvider + Sync,
 {
     // No Evil JSON Here!
-    process_minidump_with_options(dump, symbol_provider, ProcessorOptions::default())
+    process_minidump_with_options(dump, symbol_provider, ProcessorOptions::default()).await
 }
 
-/// The same as [`process_minidump`] but with extra options.
-pub fn process_minidump_with_options<'a, T, P>(
+/// Process `dump` with the given options and return a report as a `ProcessState`.
+///
+/// See [`ProcessorOptions`][] for details on the specific features that can be
+/// enabled and how to choose them.
+pub async fn process_minidump_with_options<'a, T, P>(
     dump: &Minidump<'a, T>,
     symbol_provider: &P,
-    options: ProcessorOptions,
+    options: ProcessorOptions<'_>,
 ) -> Result<ProcessState, ProcessError>
 where
     T: Deref<Target = [u8]> + 'a,
-    P: SymbolProvider,
+    P: SymbolProvider + Sync,
 {
+    options.check_deprecated_and_disabled();
+
     // Thread list is required for processing.
     let thread_list = dump
         .get_stream::<MinidumpThreadList>()
@@ -99,16 +235,7 @@ where
         .get_stream::<MinidumpSystemInfo>()
         .or(Err(ProcessError::MissingSystemInfo))?;
 
-    let mut os_version = format!(
-        "{}.{}.{}",
-        dump_system_info.raw.major_version,
-        dump_system_info.raw.minor_version,
-        dump_system_info.raw.build_number
-    );
-    if let Some(csd_version) = dump_system_info.csd_version() {
-        os_version.push(' ');
-        os_version.push_str(&csd_version);
-    }
+    let (os_version, os_build) = dump_system_info.os_parts();
 
     let linux_standard_base = dump.get_stream::<MinidumpLinuxLsbRelease>().ok();
     let linux_cpu_info = dump
@@ -162,6 +289,7 @@ where
     let system_info = SystemInfo {
         os: dump_system_info.os,
         os_version: Some(os_version),
+        os_build,
         cpu: dump_system_info.cpu,
         cpu_info,
         cpu_microcode_version,
@@ -196,7 +324,7 @@ where
     let (crash_reason, crash_address, crashing_thread_id) = if let Some(exception) = exception_ref {
         (
             Some(exception.get_crash_reason(system_info.os, system_info.cpu)),
-            Some(exception.get_crash_address(system_info.os)),
+            Some(exception.get_crash_address(system_info.os, system_info.cpu)),
             Some(exception.get_crashing_thread_id()),
         )
     } else {
@@ -230,9 +358,11 @@ where
     let mut threads = vec![];
     let mut requesting_thread = None;
     for (i, thread) in thread_list.threads.iter().enumerate() {
+        let id = thread.raw.thread_id;
+
         // If this is the thread that wrote the dump, skip processing it.
-        if dump_thread_id.is_some() && dump_thread_id.unwrap() == thread.raw.thread_id {
-            threads.push(CallStack::with_info(CallStackInfo::DumpThreadSkipped));
+        if dump_thread_id.is_some() && dump_thread_id.unwrap() == id {
+            threads.push(CallStack::with_info(id, CallStackInfo::DumpThreadSkipped));
             continue;
         }
 
@@ -246,17 +376,39 @@ where
             .unwrap_or(false)
         {
             requesting_thread = Some(i);
-            exception_context
-                .as_deref()
-                .or_else(|| thread_context.as_deref())
+            exception_context.as_deref().or(thread_context.as_deref())
         } else {
             thread_context.as_deref()
         };
 
-        let stack = thread.stack_memory(&memory_list);
+        let stack_memory = thread.stack_memory(&memory_list);
 
-        let mut stack =
-            stackwalker::walk_stack(&context, stack.as_deref(), &modules, symbol_provider);
+        let mut stack = stackwalker::walk_stack(
+            &context,
+            stack_memory.as_deref(),
+            &modules,
+            &system_info,
+            symbol_provider,
+        )
+        .await;
+        stack.thread_id = id;
+        for frame in &mut stack.frames {
+            // If the frame doesn't have a loaded module, try to find an unloaded module
+            // that overlaps with its address range. The may be multiple, so record all
+            // of them and the offsets this frame has in them.
+            if frame.module.is_none() {
+                let mut offsets = BTreeMap::new();
+                for unloaded in unloaded_modules.modules_at_address(frame.instruction) {
+                    let offset = frame.instruction - unloaded.raw.base_of_image;
+                    offsets
+                        .entry(unloaded.name.clone())
+                        .or_insert_with(BTreeSet::new)
+                        .insert(offset);
+                }
+
+                frame.unloaded_modules = offsets;
+            }
+        }
 
         let name = thread_names
             .get_name(thread.raw.thread_id)
@@ -266,6 +418,10 @@ where
 
         stack.last_error_value = thread.last_error(system_info.cpu, &memory_list);
 
+        if options.recover_function_args {
+            arg_recovery::fill_arguments(&mut stack, stack_memory.as_deref());
+        }
+
         threads.push(stack);
     }
 
@@ -278,7 +434,7 @@ where
 
     Ok(ProcessState {
         process_id,
-        time: Utc.timestamp(dump.header.time_date_stamp as i64, 0),
+        time: SystemTime::UNIX_EPOCH + Duration::from_secs(dump.header.time_date_stamp as u64),
         process_create_time,
         cert_info: evil.certs,
         crash_reason,
