@@ -9,9 +9,10 @@
 use crate::process_state::{FrameTrust, StackFrame};
 use crate::stackwalker::unwind::Unwind;
 use crate::stackwalker::CfiStackWalker;
-use crate::SymbolProvider;
+use crate::{SymbolProvider, SystemInfo};
 use log::trace;
 use minidump::format::CONTEXT_AMD64;
+use minidump::system_info::Os;
 use minidump::{
     MinidumpContext, MinidumpContextValidity, MinidumpMemory, MinidumpModuleList,
     MinidumpRawContext,
@@ -26,16 +27,16 @@ const FRAME_POINTER_REGISTER: &str = "rbp";
 // FIXME: rdi and rsi are also preserved on windows (but not in sysv) -- we should handle that?
 const CALLEE_SAVED_REGS: &[&str] = &["rbx", "rbp", "r12", "r13", "r14", "r15"];
 
-fn get_caller_by_cfi<P>(
+async fn get_caller_by_cfi<P>(
     ctx: &CONTEXT_AMD64,
     callee: &StackFrame,
     grand_callee: Option<&StackFrame>,
-    stack_memory: &MinidumpMemory,
+    stack_memory: &MinidumpMemory<'_>,
     modules: &MinidumpModuleList,
     symbol_provider: &P,
 ) -> Option<StackFrame>
 where
-    P: SymbolProvider,
+    P: SymbolProvider + Sync,
 {
     trace!("unwind: trying cfi");
 
@@ -49,9 +50,11 @@ where
     let module = modules.module_at_address(callee.instruction)?;
 
     let grand_callee_parameter_size = grand_callee.and_then(|f| f.parameter_size).unwrap_or(0);
+    let has_grand_callee = grand_callee.is_some();
 
     let mut stack_walker = CfiStackWalker {
         instruction: callee.instruction,
+        has_grand_callee,
         grand_callee_parameter_size,
 
         callee_ctx: ctx,
@@ -66,7 +69,9 @@ where
         stack_memory,
     };
 
-    symbol_provider.walk_frame(module, &mut stack_walker)?;
+    symbol_provider
+        .walk_frame(module, &mut stack_walker)
+        .await?;
     let caller_ip = stack_walker.caller_ctx.rip;
     let caller_sp = stack_walker.caller_ctx.rsp;
 
@@ -104,13 +109,26 @@ fn callee_forwarded_regs(valid: &MinidumpContextValidity) -> HashSet<&'static st
 fn get_caller_by_frame_pointer<P>(
     ctx: &CONTEXT_AMD64,
     callee: &StackFrame,
-    stack_memory: &MinidumpMemory,
+    stack_memory: &MinidumpMemory<'_>,
     _modules: &MinidumpModuleList,
+    system_info: &SystemInfo,
     _symbol_provider: &P,
 ) -> Option<StackFrame>
 where
-    P: SymbolProvider,
+    P: SymbolProvider + Sync,
 {
+    // On Windows x64, frame-pointer unwinding purely with the data on the stack
+    // is not possible, as proper unwinding requires access to `UNWIND_INFO`,
+    // because the frame pointer does not necessarily point to the end of the
+    // frame.
+    // In particular, the docs state that:
+    // > [The frame register] offset permits pointing the FP register into the
+    // > middle of the local stack allocation [...]
+    // https://docs.microsoft.com/en-us/cpp/build/exception-handling-x64
+    if system_info.os == Os::Windows {
+        return None;
+    }
+
     trace!("unwind: trying frame pointer");
     if let MinidumpContextValidity::Some(ref which) = callee.context.valid {
         if !which.contains(FRAME_POINTER_REGISTER) {
@@ -202,15 +220,15 @@ where
     Some(StackFrame::from_context(context, FrameTrust::FramePointer))
 }
 
-fn get_caller_by_scan<P>(
+async fn get_caller_by_scan<P>(
     ctx: &CONTEXT_AMD64,
     callee: &StackFrame,
-    stack_memory: &MinidumpMemory,
+    stack_memory: &MinidumpMemory<'_>,
     modules: &MinidumpModuleList,
     symbol_provider: &P,
 ) -> Option<StackFrame>
 where
-    P: SymbolProvider,
+    P: SymbolProvider + Sync,
 {
     trace!("unwind: trying scan");
     // Stack scanning is just walking from the end of the frame until we encounter
@@ -250,7 +268,7 @@ where
     for i in 0..scan_range {
         let address_of_ip = last_sp.checked_add(i * POINTER_WIDTH)?;
         let caller_ip = stack_memory.get_memory_at_address(address_of_ip as u64)?;
-        if instruction_seems_valid(caller_ip, modules, symbol_provider) {
+        if instruction_seems_valid(caller_ip, modules, symbol_provider).await {
             // ip is pushed by CALL, so sp is just address_of_ip + ptr
             let caller_sp = address_of_ip.checked_add(POINTER_WIDTH)?;
 
@@ -279,24 +297,31 @@ where
             // It's unclear why, perhaps the test is buggy, but for now we
             // preserve that behaviour.
             if let Some(last_bp) = last_bp {
-                let address_of_bp = address_of_ip - POINTER_WIDTH;
-                // Can assume this resolves because we already walked over it when
-                // checking address_of_ip values.
-                let bp = stack_memory.get_memory_at_address(address_of_bp as u64)?;
-                if last_bp == address_of_bp
-                    && bp > address_of_ip
-                    && bp - address_of_bp <= MAX_REASONABLE_GAP_BETWEEN_FRAMES
-                {
-                    // Final sanity check that resulting bp is still inside stack memory.
-                    if stack_memory
-                        .get_memory_at_address::<Pointer>(bp as u64)
-                        .is_some()
+                // If we're on the first iteration of the scan, there can't possibly be a
+                // frame pointer, because the entire stack frame is taken up by the return
+                // pointer. And if we're not on the first iteration, then the last iteration
+                // already loaded the location we expect the frame pointer to be in, so we can
+                // unconditionally load it here.
+                if i > 0 {
+                    let address_of_bp = address_of_ip - POINTER_WIDTH;
+                    // Can assume this resolves because we already walked over it when
+                    // checking address_of_ip values.
+                    let bp = stack_memory.get_memory_at_address(address_of_bp as u64)?;
+                    if last_bp == address_of_bp
+                        && bp > address_of_ip
+                        && bp - address_of_bp <= MAX_REASONABLE_GAP_BETWEEN_FRAMES
                     {
-                        caller_bp = Some(bp);
+                        // Final sanity check that resulting bp is still inside stack memory.
+                        if stack_memory
+                            .get_memory_at_address::<Pointer>(bp as u64)
+                            .is_some()
+                        {
+                            caller_bp = Some(bp);
+                        }
+                    } else if last_bp >= caller_sp {
+                        // Don't sanity check that the address is inside the stack? Hmm.
+                        caller_bp = Some(last_bp);
                     }
-                } else if last_bp >= caller_sp {
-                    // Don't sanity check that the address is inside the stack? Hmm.
-                    caller_bp = Some(last_bp);
                 }
             }
 
@@ -352,25 +377,25 @@ where
 /// If we applied this more rigorous validation to cfi/fp methods, we
 /// would just discard the correct register values from the known frame
 /// and immediately start doing unreliable scans.
-fn instruction_seems_valid<P>(
+async fn instruction_seems_valid<P>(
     instruction: Pointer,
     modules: &MinidumpModuleList,
     symbol_provider: &P,
 ) -> bool
 where
-    P: SymbolProvider,
+    P: SymbolProvider + Sync,
 {
     if is_non_canonical(instruction) || instruction == 0 {
         return false;
     }
 
-    super::instruction_seems_valid_by_symbols(instruction as u64, modules, symbol_provider)
+    super::instruction_seems_valid_by_symbols(instruction as u64, modules, symbol_provider).await
 }
 
 fn stack_seems_valid(
     caller_sp: Pointer,
     callee_sp: Pointer,
-    stack_memory: &MinidumpMemory,
+    stack_memory: &MinidumpMemory<'_>,
 ) -> bool {
     // The stack shouldn't *grow* when we unwind
     if caller_sp <= callee_sp {
@@ -399,55 +424,63 @@ fn is_non_canonical(ptr: Pointer) -> bool {
     ptr > 0x7FFFFFFFFFFF && ptr < 0xFFFF800000000000
 }
 
+#[async_trait::async_trait]
 impl Unwind for CONTEXT_AMD64 {
-    fn get_caller_frame<P>(
+    async fn get_caller_frame<P>(
         &self,
         callee: &StackFrame,
         grand_callee: Option<&StackFrame>,
-        stack_memory: Option<&MinidumpMemory>,
+        stack_memory: Option<&MinidumpMemory<'_>>,
         modules: &MinidumpModuleList,
+        system_info: &SystemInfo,
         syms: &P,
     ) -> Option<StackFrame>
     where
-        P: SymbolProvider,
+        P: SymbolProvider + Sync,
     {
-        stack_memory
-            .as_ref()
-            .and_then(|stack| {
-                get_caller_by_cfi(self, callee, grand_callee, stack, modules, syms)
-                    .or_else(|| get_caller_by_frame_pointer(self, callee, stack, modules, syms))
-                    .or_else(|| get_caller_by_scan(self, callee, stack, modules, syms))
-            })
-            .and_then(|mut frame| {
-                // We now check the frame to see if it looks like unwinding is complete,
-                // based on the frame we computed having a nonsense value. Returning
-                // None signals to the unwinder to stop unwinding.
+        let stack = stack_memory.as_ref()?;
 
-                // if the instruction is within the first ~page of memory, it's basically
-                // null, and we can assume unwinding is complete.
-                if frame.context.get_instruction_pointer() < 4096 {
-                    trace!("unwind: instruction pointer was nullish, assuming unwind complete");
-                    return None;
-                }
-                // If the new stack pointer is at a lower address than the old,
-                // then that's clearly incorrect. Treat this as end-of-stack to
-                // enforce progress and avoid infinite loops.
-                if frame.context.get_stack_pointer() <= self.rsp {
-                    trace!("unwind: stack pointer went backwards, assuming unwind complete");
-                    return None;
-                }
+        // .await doesn't like closures, so don't use Option chaining
+        let mut frame = None;
+        if frame.is_none() {
+            frame = get_caller_by_cfi(self, callee, grand_callee, stack, modules, syms).await;
+        }
+        if frame.is_none() {
+            frame = get_caller_by_frame_pointer(self, callee, stack, modules, system_info, syms);
+        }
+        if frame.is_none() {
+            frame = get_caller_by_scan(self, callee, stack, modules, syms).await;
+        }
+        let mut frame = frame?;
 
-                // Ok, the frame now seems well and truly valid, do final cleanup.
+        // We now check the frame to see if it looks like unwinding is complete,
+        // based on the frame we computed having a nonsense value. Returning
+        // None signals to the unwinder to stop unwinding.
 
-                // A caller's ip is the return address, which is the instruction
-                // *after* the CALL that caused us to arrive at the callee. Set
-                // the value to one less than that, so it points within the
-                // CALL instruction. This is important because we use this value
-                // to lookup the CFI we need to unwind the next frame.
-                let ip = frame.context.get_instruction_pointer() as u64;
-                frame.instruction = ip - 1;
+        // if the instruction is within the first ~page of memory, it's basically
+        // null, and we can assume unwinding is complete.
+        if frame.context.get_instruction_pointer() < 4096 {
+            trace!("unwind: instruction pointer was nullish, assuming unwind complete");
+            return None;
+        }
+        // If the new stack pointer is at a lower address than the old,
+        // then that's clearly incorrect. Treat this as end-of-stack to
+        // enforce progress and avoid infinite loops.
+        if frame.context.get_stack_pointer() <= self.rsp {
+            trace!("unwind: stack pointer went backwards, assuming unwind complete");
+            return None;
+        }
 
-                Some(frame)
-            })
+        // Ok, the frame now seems well and truly valid, do final cleanup.
+
+        // A caller's ip is the return address, which is the instruction
+        // *after* the CALL that caused us to arrive at the callee. Set
+        // the value to one less than that, so it points within the
+        // CALL instruction. This is important because we use this value
+        // to lookup the CFI we need to unwind the next frame.
+        let ip = frame.context.get_instruction_pointer() as u64;
+        frame.instruction = ip - 1;
+
+        Some(frame)
     }
 }
