@@ -3,8 +3,9 @@
 
 use crate::process_state::*;
 use crate::stackwalker::walk_stack;
-use crate::{string_symbol_supplier, Symbolizer};
+use crate::{string_symbol_supplier, Symbolizer, SystemInfo};
 use minidump::format::CONTEXT_AMD64;
+use minidump::system_info::{Cpu, Os};
 use minidump::*;
 use std::collections::HashMap;
 use test_assembler::*;
@@ -12,6 +13,7 @@ use test_assembler::*;
 struct TestFixture {
     pub raw: CONTEXT_AMD64,
     pub modules: MinidumpModuleList,
+    pub system_info: SystemInfo,
     pub symbols: HashMap<String, String>,
 }
 
@@ -25,11 +27,20 @@ impl TestFixture {
                 MinidumpModule::new(0x00007400c0000000, 0x10000, "module1"),
                 MinidumpModule::new(0x00007500b0000000, 0x10000, "module2"),
             ]),
+            system_info: SystemInfo {
+                os: Os::Linux,
+                os_version: None,
+                os_build: None,
+                cpu: Cpu::X86_64,
+                cpu_info: None,
+                cpu_microcode_version: None,
+                cpu_count: 1,
+            },
             symbols: HashMap::new(),
         }
     }
 
-    pub fn walk_stack(&self, stack: Section) -> CallStack {
+    pub async fn walk_stack(&self, stack: Section) -> CallStack {
         let context = MinidumpContext {
             raw: MinidumpRawContext::Amd64(self.raw.clone()),
             valid: MinidumpContextValidity::All,
@@ -48,8 +59,10 @@ impl TestFixture {
             &Some(&context),
             Some(&stack_memory),
             &self.modules,
+            &self.system_info,
             &symbolizer,
         )
+        .await
     }
 
     pub fn add_symbols(&mut self, name: String, symbols: String) {
@@ -57,8 +70,8 @@ impl TestFixture {
     }
 }
 
-#[test]
-fn test_simple() {
+#[tokio::test]
+async fn test_simple() {
     let mut f = TestFixture::new();
     let stack = Section::new();
     stack.start().set_const(0x80000000);
@@ -69,15 +82,15 @@ fn test_simple() {
     f.raw.rip = 0x00007400c0000200;
     f.raw.rbp = 0x8000000080000000;
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 1);
     let f = &s.frames[0];
     let m = f.module.as_ref().unwrap();
     assert_eq!(m.code_file(), "module1");
 }
 
-#[test]
-fn test_caller_pushed_rbp() {
+#[tokio::test]
+async fn test_caller_pushed_rbp() {
     // Functions typically push their %rbp upon entry and set %rbp pointing
     // there.  If stackwalking finds a plausible address for the next frame's
     // %rbp directly below the return address, assume that it is indeed the
@@ -112,7 +125,7 @@ fn test_caller_pushed_rbp() {
     f.raw.rbp = frame0_rbp.value().unwrap();
     f.raw.rsp = stack.start().value().unwrap();
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -148,8 +161,76 @@ fn test_caller_pushed_rbp() {
     }
 }
 
-#[test]
-fn test_scan_without_symbols() {
+#[tokio::test]
+async fn test_windows_rbp_scan() {
+    let mut f = TestFixture::new();
+    f.system_info.os = Os::Windows;
+
+    let mut stack = Section::new();
+    let stack_start = 0x8000000080000000;
+    let return_address = 0x00007500b0000110;
+    stack.start().set_const(stack_start);
+
+    let frame0_rbp = Label::new();
+    let frame1_sp = Label::new();
+    let frame1_rbp = Label::new();
+
+    stack = stack
+        // frame 0
+        .append_repeated(0, 16) // space
+        .D64(0x00000000b0000000) // junk that's not
+        .D64(0x00000000b0000000) // a return address
+        .mark(&frame0_rbp) // the FP can point to the middle of the stack on Windows
+        .D64(0x00000000c0001000)
+        .D64(0x00000000b000aaaa)
+        .D64(&frame1_rbp) // caller-pushed %rbp
+        .D64(return_address) // actual return address
+        // frame 1
+        .mark(&frame1_sp)
+        .append_repeated(0, 32) // body of frame1
+        .mark(&frame1_rbp) // end of stack
+        .D64(0);
+
+    f.raw.rip = 0x00007400c0000200;
+    f.raw.rbp = frame0_rbp.value().unwrap();
+    f.raw.rsp = stack.start().value().unwrap();
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 2);
+
+    {
+        // To avoid reusing locals by mistake
+        let f0 = &s.frames[0];
+        assert_eq!(f0.trust, FrameTrust::Context);
+        assert_eq!(f0.context.valid, MinidumpContextValidity::All);
+        if let MinidumpRawContext::Amd64(ctx) = &f0.context.raw {
+            assert_eq!(ctx.rbp, frame0_rbp.value().unwrap());
+        } else {
+            unreachable!();
+        }
+    }
+
+    {
+        // To avoid reusing locals by mistake
+        let f1 = &s.frames[1];
+        assert_eq!(f1.trust, FrameTrust::Scan);
+        if let MinidumpContextValidity::Some(ref which) = f1.context.valid {
+            assert!(which.contains("rip"));
+            assert!(which.contains("rsp"));
+        } else {
+            unreachable!();
+        }
+        if let MinidumpRawContext::Amd64(ctx) = &f1.context.raw {
+            assert_eq!(ctx.rip, return_address);
+            assert_eq!(ctx.rsp, frame1_sp.value().unwrap());
+        } else {
+            unreachable!();
+        }
+    }
+}
+
+#[tokio::test]
+async fn test_scan_without_symbols() {
     // When the stack walker resorts to scanning the stack,
     // only addresses located within loaded modules are
     // considered valid return addresses.
@@ -190,7 +271,7 @@ fn test_scan_without_symbols() {
     f.raw.rbp = frame1_rbp.value().unwrap();
     f.raw.rsp = stack.start().value().unwrap();
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 3);
 
     {
@@ -241,8 +322,8 @@ fn test_scan_without_symbols() {
     }
 }
 
-#[test]
-fn test_scan_with_symbols() {
+#[tokio::test]
+async fn test_scan_with_symbols() {
     // Test that we can refine our scanning using symbols. Specifically we
     // should be able to reject pointers that are in modules but don't map to
     // any FUNC/PUBLIC record.
@@ -283,7 +364,7 @@ fn test_scan_with_symbols() {
         String::from("FUNC 100 400 10 marsupial\n"),
     );
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -369,13 +450,13 @@ fn init_cfi_state() -> (TestFixture, Section, CONTEXT_AMD64, MinidumpContextVali
     (f, stack, expected, expected_valid)
 }
 
-fn check_cfi(
+async fn check_cfi(
     f: TestFixture,
     stack: Section,
     expected: CONTEXT_AMD64,
     expected_valid: MinidumpContextValidity,
 ) {
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -413,8 +494,8 @@ fn check_cfi(
     unreachable!();
 }
 
-#[test]
-fn test_cfi_at_4000() {
+#[tokio::test]
+async fn test_cfi_at_4000() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -426,11 +507,11 @@ fn test_cfi_at_4000() {
     expected.set_register("rsp", frame1_rsp.value().unwrap());
     f.raw.set_register("rip", 0x00007400c0004000);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4001() {
+#[tokio::test]
+async fn test_cfi_at_4001() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -444,11 +525,11 @@ fn test_cfi_at_4001() {
     f.raw.set_register("rip", 0x00007400c0004001);
     f.raw.set_register("rbx", 0xbe0487d2f9eafe29);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4002() {
+#[tokio::test]
+async fn test_cfi_at_4002() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -463,11 +544,11 @@ fn test_cfi_at_4002() {
     f.raw.set_register("rbx", 0xed1b02e8cc0fc79c); // saved %r12
     f.raw.set_register("r12", 0xb0118de918a4bcea); // callee's (distinct) %r12 value
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4003() {
+#[tokio::test]
+async fn test_cfi_at_4003() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -486,11 +567,11 @@ fn test_cfi_at_4003() {
     f.raw.set_register("r12", 0x89d04fa804c87a43); // callee's (distinct) %r12
     f.raw.set_register("r13", 0x5118e02cbdb24b03); // callee's (distinct) %r13
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4004() {
+#[tokio::test]
+async fn test_cfi_at_4004() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -509,11 +590,11 @@ fn test_cfi_at_4004() {
     f.raw.set_register("r12", 0x46b1b8868891b34a); // callee's (distinct) %r12
     f.raw.set_register("r13", 0x5118e02cbdb24b03); // callee's (distinct) %r13
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4005() {
+#[tokio::test]
+async fn test_cfi_at_4005() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -532,11 +613,11 @@ fn test_cfi_at_4005() {
     f.raw.set_register("r12", 0x46b1b8868891b34a); // callee's %r12
     f.raw.set_register("r13", 0x00007400c0005510); // return address
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4006() {
+#[tokio::test]
+async fn test_cfi_at_4006() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame0_rbp = Label::new();
@@ -558,33 +639,169 @@ fn test_cfi_at_4006() {
     f.raw.set_register("r12", 0x26e007b341acfebd); // callee's %r12
     f.raw.set_register("r13", 0x00007400c0005510); // return address
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_overflow() {
+#[tokio::test]
+async fn test_frame_pointer_overflow() {
     // Make sure we don't explode when trying frame pointer analysis on a value
     // that will overflow.
 
-    // Functions typically push their %rbp upon entry and set %rbp pointing
-    // there.  If stackwalking finds a plausible address for the next frame's
-    // %rbp directly below the return address, assume that it is indeed the
-    // next frame's %rbp.
+    type Pointer = u64;
+    let stack_max: Pointer = Pointer::MAX;
+    let stack_size: Pointer = 1000;
+    let bad_frame_ptr: Pointer = stack_max;
+
     let mut f = TestFixture::new();
     let mut stack = Section::new();
-    let stack_start = 0x8000000080000000;
-    stack.start().set_const(stack_start);
+    let stack_start: Pointer = stack_max - stack_size;
+    stack.start().set_const(stack_start as u64);
 
     stack = stack
         // frame 0
-        .append_repeated(0, 1000); // junk, not important to the test
+        .append_repeated(0, stack_size as usize); // junk, not important to the test
 
     f.raw.rip = 0x00007400c0000200;
-    f.raw.rbp = u64::MAX;
-    f.raw.rsp = stack.start().value().unwrap();
+    f.raw.rbp = bad_frame_ptr;
+    f.raw.rsp = stack.start().value().unwrap() as Pointer;
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 1);
 
     // As long as we don't panic, we're good!
 }
+
+#[tokio::test]
+async fn test_frame_pointer_barely_no_overflow() {
+    // This is test_caller_pushed_rbp but with the all the values pushed
+    // as close to the upper memory boundary as possible, to confirm that
+    // our code doesn't randomly overflow *AND* isn't overzealous in
+    // its overflow guards.
+
+    let mut f = TestFixture::new();
+    let mut stack = Section::new();
+
+    type Pointer = u64;
+    let stack_max: Pointer = Pointer::MAX;
+    let pointer_size: Pointer = std::mem::size_of::<Pointer>() as Pointer;
+    let stack_size: Pointer = pointer_size * 3;
+
+    let stack_start: Pointer = stack_max - stack_size;
+    let return_address: Pointer = 0x00007500b0000110;
+    stack.start().set_const(stack_start as u64);
+
+    let frame0_fp = Label::new();
+    let frame1_sp = Label::new();
+    let frame1_fp = Label::new();
+
+    stack = stack
+        // frame 0
+        .mark(&frame0_fp)
+        .D64(&frame1_fp) // caller-pushed %rbp
+        .D64(return_address) // actual return address
+        // frame 1
+        .mark(&frame1_sp)
+        .mark(&frame1_fp) // end of stack
+        .D64(0);
+
+    f.raw.rip = 0x00007400c0000200;
+    f.raw.rbp = frame0_fp.value().unwrap() as Pointer;
+    f.raw.rsp = stack.start().value().unwrap();
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 2);
+
+    {
+        // To avoid reusing locals by mistake
+        let f0 = &s.frames[0];
+        assert_eq!(f0.trust, FrameTrust::Context);
+        assert_eq!(f0.context.valid, MinidumpContextValidity::All);
+        if let MinidumpRawContext::Amd64(ctx) = &f0.context.raw {
+            assert_eq!(ctx.rbp, frame0_fp.value().unwrap() as Pointer);
+        } else {
+            unreachable!();
+        }
+    }
+
+    {
+        // To avoid reusing locals by mistake
+        let f1 = &s.frames[1];
+        assert_eq!(f1.trust, FrameTrust::FramePointer);
+        if let MinidumpContextValidity::Some(ref which) = f1.context.valid {
+            assert!(which.contains("rip"));
+            assert!(which.contains("rsp"));
+            assert!(which.contains("rbp"));
+        } else {
+            unreachable!();
+        }
+        if let MinidumpRawContext::Amd64(ctx) = &f1.context.raw {
+            assert_eq!(ctx.rip, return_address);
+            assert_eq!(ctx.rsp, frame1_sp.value().unwrap() as Pointer);
+            assert_eq!(ctx.rbp, frame1_fp.value().unwrap() as Pointer);
+        } else {
+            unreachable!();
+        }
+    }
+}
+
+#[tokio::test]
+async fn test_scan_walk_overflow() {
+    // There's a possible overflow when address_of_ip starts out at 0.
+    //
+    // To avoid this, we only try to recover rbp when we're scanning at least
+    // 1 pointer width away from the start of the stack.
+    let mut f = TestFixture::new();
+    let mut stack = Section::new();
+    let stack_start = 0;
+    stack.start().set_const(stack_start);
+
+    let return_address1 = 0x00007500b0000100_u64;
+
+    let frame1_sp = Label::new();
+    let frame1_rbp = Label::new();
+
+    stack = stack
+        // frame 0
+        .D64(return_address1) // actual return address
+        // frame 1
+        .mark(&frame1_sp)
+        .append_repeated(0, 16) // space
+        .D64(0x00007400b0000000) // more junk
+        .D64(0x00007500d0000000)
+        .mark(&frame1_rbp);
+
+    f.raw.rip = 0x00007400c0000200;
+    f.raw.rbp = frame1_rbp.value().unwrap();
+    f.raw.rsp = stack.start().value().unwrap();
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 2);
+
+    {
+        // To avoid reusing locals by mistake
+        let f0 = &s.frames[0];
+        assert_eq!(f0.trust, FrameTrust::Context);
+        assert_eq!(f0.context.valid, MinidumpContextValidity::All);
+    }
+
+    {
+        // To avoid reusing locals by mistake
+        let f1 = &s.frames[1];
+        assert_eq!(f1.trust, FrameTrust::Scan);
+        if let MinidumpContextValidity::Some(ref which) = f1.context.valid {
+            assert!(which.contains("rip"));
+            assert!(which.contains("rsp"));
+        } else {
+            unreachable!();
+        }
+
+        if let MinidumpRawContext::Amd64(ctx) = &f1.context.raw {
+            assert_eq!(ctx.rip, return_address1);
+            assert_eq!(ctx.rsp, frame1_sp.value().unwrap());
+            // We were unable to recover rbp, so it defaulted to 0.
+            assert_eq!(ctx.rbp, 0);
+        } else {
+            unreachable!();
+        }
+    }
+}
