@@ -2,20 +2,22 @@
 // file at the top-level directory of this distribution.
 
 use crate::process_state::StackFrame;
-use crate::SymbolProvider;
+use crate::{SymbolProvider, SystemInfo};
 use minidump::{MinidumpMemory, MinidumpModuleList};
 
 /// A trait for things that can unwind to a caller.
+#[async_trait::async_trait]
 pub trait Unwind {
     /// Get the caller frame of this frame.
-    fn get_caller_frame<P>(
+    async fn get_caller_frame<P>(
         &self,
         callee: &StackFrame,
         grand_callee: Option<&StackFrame>,
-        stack_memory: Option<&MinidumpMemory>,
+        stack_memory: Option<&MinidumpMemory<'_>>,
         modules: &MinidumpModuleList,
+        system_info: &SystemInfo,
         symbol_provider: &P,
     ) -> Option<StackFrame>
     where
-        P: SymbolProvider;
+        P: SymbolProvider + Sync;
 }
